#!/usr/bin/env python3

from argparse import ArgumentParser, ArgumentTypeError, \
                     RawDescriptionHelpFormatter, BooleanOptionalAction
import sys
import os
import re
import pickle
from operator import add
from functools import reduce
from itertools import product
import h5py
import numpy as np
import matplotlib.pyplot as plt

import gaussian_suite.gen_input as gsgi
import xyz_py as xyzp
from xyz_py import atomic
from molcas_suite.cli import ParseExtra
import gaussian_suite.cd_extractor as gscd
import hpc_suite as hpc
from hpc_suite.action import ParseKwargs, OrderedSelectionAction
from hpc_suite import make_parse_dict, SecondaryHelp
import angmom_suite
from angmom_suite import crystal
from angmom_suite.basis import Level
from angmom_suite.multi_electron import Ion
import molvis.core as mvis
from pymatgen.core.structure import Molecule, Structure
from pymatgen.transformations.site_transformations import TranslateSitesTransformation

from .coordinates import CoordinateInfo
from .derivative import print_tau_style, read_tau_style
from .lvc import LVC, generate_lvc_input, make_lvc_evaluator, ANG2BOHR, \
    LVCModelHamiltonian, LVCAdiabaticEnergies, LVCDiabaticHamiltonian
from .vibrations import Harmonic
from . import cells

plt.rcParams['font.family'] = "Arial"

HARTREE2INVCM = 219474.63 


# Action for secondary function help message
class FunctionHelp(SecondaryHelp):
    def __call__(self, parser, namespace, values, option_string=None):

        if namespace.function == 'cfp':
            angmom_suite.read_args(['cfp', '--help'])
        else:
            raise ValueError("Supply valid option to function argument.")


def str_int(x):
    try:
        return int(x)
    except ValueError:
        return str(x)


def coords_func(args):
    sp_data = CoordinateInfo.from_args(args)
    sp_data.to_file(args.distortion_info)

    return


def lvc_func(args):

    if args.add is not None:
        def join(a, b):
            return a.join(b, xyzp.load_xyz(args.reference_xyz)[1] * ANG2BOHR)
        lvc = reduce(join, (LVC.from_file(f_lvc) for f_lvc in args.add))

    elif args.efld is not None:
        charge_list = [np.loadtxt(f_chrg) for f_chrg in args.charge]
        xyz_list = [xyzp.load_xyz(f_xyz)[1] * ANG2BOHR for f_xyz in args.xyz]
        lvc = LVC.from_molcas_frozen_density(
            args.rassi, args.efld, charge_list, xyz_list)

    elif args.grad is not None:
        lvc = LVC.from_molcas(args.rassi, args.grad)

    else:
        raise ValueError("LVC gradient information supplied!")

    if args.active_atoms is not None:
        idc = [idx - 1 for idc in args.active_atoms for idx in idc]
        lvc = lvc.subset_atoms(active_atom_idc=idc)

    elif args.ref_coordinates is not None:
        _, coords = xyzp.load_xyz(args.ref_coordinates)
        lvc = lvc.subset_atoms(ref_coords=coords * ANG2BOHR)

    lvc.to_file(args.lvc_data, verbose=args.verbose)

    return


def eval_func(args, unknown_args):

    try:
        selected = args._selection
    except AttributeError:
        sys.exit("No quantity selected for evaluation!")

    # Resolve coordinates
    if args.vibration_info is not None:
        ho = Harmonic.from_file(args.vibration_info)
        coords = ho.mass_freq_weighted_coordinates * ANG2BOHR
    else:
        coords = None

    store_args = hpc.read_args(['store'] + unknown_args)

    eval_args = (store_args, make_lvc_evaluator, selected)
    eval_kwargs = {"order": args.order, "coords": coords,
                   "truncate": args.truncate, "align": args.align}

    if args.geom is None:
        hpc.store_func(*eval_args, geom=None, **eval_kwargs)
    else:
        for geom in map(lambda file: xyzp.load_xyz(file)[1], args.geom):
            hpc.store_func(*eval_args, geom=geom * ANG2BOHR, **eval_kwargs)


def vib_func(args):

    if args.gaussian_log:

        if args.mass:
            raise ValueError("The --mass argument is not available with freq "
                             "data read from Gaussian text output use *.fchk!")

        ho = Harmonic.from_gaussian_log(args.gaussian_log)

    elif args.gaussian_fchk:
        ho = Harmonic.from_gaussian_fchk(
            args.gaussian_fchk,
            ext_masses=args.mass,
            rot=args.rot,
            trans=args.trans
        )

    elif args.force_sets:

        ho = Harmonic.from_vasp_phonopy(
            args.poscar,
            args.force_sets,
            ext_masses=args.mass,
            trans=args.trans,
            distortion_expansion=args.distortion_expansion,
            distortion_cutoff=args.distortion_cutoff,
            central_index=args.central_index,
            force_expansion=args.force_expansion,
            q_mesh=args.q_mesh,
            gamma_center=args.gamma_center
        )

    elif args.phonopy_hdf5:

        if args.mass:
            raise ValueError("The --mass argument is not available with phonon"
                             " data read from phonopy hdf5 database, use "
                             "FORCE_SETS instead!")

        ho = Harmonic.from_phonopy_hdf5(
            args.poscar,
            args.phonopy_hdf5,
            trans=args.trans,
            distortion_expansion=args.distortion_expansion,
            distortion_cutoff=args.distortion_cutoff,
            central_index=args.central_index
        )

    if args.active_atoms is not None:
        idc = [idx - 1 for idc in args.active_atoms for idx in idc]
        ho = ho.subset_atoms(active_atom_idc=idc)

    elif args.ref_coordinates is not None:
        _, coords = xyzp.load_xyz(args.ref_coordinates)
        ho = ho.subset_atoms(ref_coords=coords)

    ho.to_file(args.vibration_info, store_vecs=args.store_vecs)

    if args.save_pyvibms is not None:
        ho.to_pyvibms(args.save_pyvibms)

    return


def prepare_func(args):

    if args.program == 'tau':
        # CFP args
        proj_kwargs = {
            'model_space': args.ground_level,
            'quax': args.quax,
            'terms': {"cf": [('J',)]},
            'k_max': args.k_max,
            'theta': True,
            'ion': args.ion
        }

        # EQ_CFPs.dat
        cfp_eq = LVCModelHamiltonian(LVC.from_file(args.lvc_data), 0,
                                     **proj_kwargs)[("cf", "J")]

        cfp_data = np.column_stack(
            (np.ones(27), list(cfp_eq.keys()), list(cfp_eq.values())))

        np.savetxt('EQ_CFPs.dat', cfp_data,
                   fmt=['%2d', '%2d', '% 2d', '%16.8f'])

        # mode_energies.dat
        ho = Harmonic.from_file(args.vibration_info)
        np.savetxt('mode_energies.dat', ho.freqs, fmt='%16.8f',
                   header="Frequency (cm-1)", comments='')

        cfp_dq = LVCModelHamiltonian(
            LVC.from_file(args.lvc_data), 1,
            coords=ho.mass_freq_weighted_coordinates * ANG2BOHR,
            **proj_kwargs
        )

        print_tau_style(filter(lambda pars: pars[0][1] == 'cf', iter(cfp_dq)),
                        ho.freqs, 'CFP_polynomials.dat')

    elif args.program == 'tau_direct':

        ho = Harmonic.from_file(args.vibration_info)
        np.savetxt('mode_energies.dat', ho.freqs, fmt='%16.8f',
                   header="Frequency (cm-1)", comments='')

        lvc = LVC.from_file(args.lvc_data)

        ener = LVCAdiabaticEnergies(lvc, 0, soc=True,
                                    truncate=args.truncate)[()]

        nstates = args.truncate or ener.size

        couplings = LVCDiabaticHamiltonian(lvc, 1, 
            coords=ho.mass_freq_weighted_coordinates * ANG2BOHR,
            soc=True, truncate=nstates)

        with h5py.File('tau.hdf5', 'w') as h, open('coupling_strength.dat', 'w') as f:

            h.create_dataset('energies', data=(ener - ener[0]) * HARTREE2INVCM)

            grp = h.create_group("couplings")

            for modes, lin in iter(couplings):

                cplg = lin * HARTREE2INVCM

                if args.trace_less:
                    cplg -= np.identity(nstates) * np.trace(cplg) / nstates

                grp.create_dataset('_'.join(map(str, modes)), data=cplg)
                strength = np.sum((cplg * cplg.conj()).real)
                f.write(f'{strength}\n')

            angm_grp = h.create_group('angmom')
            spin_grp = h.create_group('spin')

            for idx, comp in enumerate(["x", "y", "z"]):
                idc = np.ix_(range(nstates), range(nstates))
                angm_grp.create_dataset(comp, data=lvc.sos_angm[idx][idc])
                spin_grp.create_dataset(comp, data=lvc.sos_spin[idx][idc])

    return


def charges_func(args):
    """
    Wrapper function for cli call to charges
    """

    structure = Structure.from_file(args.poscar)
    equivalent_positions, elements, mappings, entities = \
        cells.get_unique_entities(structure)

    if args.gen:
        gen_charges_func(entities)
    elif args.parse:
        parse_charges_func(equivalent_positions, elements, mappings, entities)
    else:
        ValueError(f"{args.mode} mode not supported!")


def format_formula(formula):
    return formula.replace(" ", "")


def gen_charges_func(entities):
    """
    Wrapper function for cli call to charges gen
    """

    print("******************************")
    print("The unique entities are")
    for mol in entities:
        print(mol.formula)
        mol.to(f"{format_formula(mol.formula)}.xyz")
    print("******************************")

    # Create gaussian inputs for CHELPG calculation, one for each
    # member of the ASU

    # All electron aug-cc-pvtz
    ae_pvtz = [
        'H', 'Li', 'Be', 'Na', 'Mg', 'B', 'C', 'N', 'O', 'F', 'He', 'Ne', 'Al',
        'Si', 'P', 'S', 'Cl', 'Ar', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr'
    ]
    ae_basis = {atom: "aug-cc-pvtz" for atom in ae_pvtz}

    unsupported = [
        "K", "Ca", "Rb", "Sr", "Cs", "Ba", "Ac", "Th", "Pa", "U", "Np",
        "Pu", "Am", "Cm"
    ]
    unsupported_basis = {atom: "????" for atom in unsupported}

    for lab in (site.specie.symbol for mol in entities for site in mol.sites):
        if lab in unsupported:
            print(f"\033[31m Warning, add basis for {lab} to .com file\033[0m \n")

    ae_basis = ae_basis | unsupported_basis

    # ECP
    pp_pvtz = [
        'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn'
    ]
    small_pp_basis = {atom: "aug-cc-pvtz-pp" for atom in pp_pvtz}

    stuttgart = {metal: "stuttgart rsc 1997"
                 for metal in atomic.lanthanides + atomic.transition_metals}

    psuedo = small_pp_basis | stuttgart

    try:
        os.mkdir("gaussian")
    except OSError:
        pass

    for idx, mol in enumerate(entities, start=1):
        gsgi.gen_input(
            f"gaussian/{format_formula(mol.formula)}_{idx}.com",
            list(map(lambda site: site.specie.symbol, mol.sites)),
            list(map(lambda site: (site.x, site.y, site.z), mol.sites)),
            99,
            99,
            method="PBE",
            bs_spec=ae_basis,
            ecp_spec=psuedo,
            opt=False,
            freq=False,
            extra_title=f"{format_formula(mol.formula)} (molecule #{idx})",
            chelpg="charge"
        )
        print()

    print("\033[93mCharge and multiplicity set to 99 in all .com files")
    print("Please replace with actual values before submitting \033[0m \n")

    return


def parse_charges_func(equivalent_positions, elements, mappings, entities):
    """
    Wrapper function for cli call to charges parse
    """

    def read_charges(mol, idx):
        file = f"gaussian/{format_formula(mol.formula)}_{idx}.log"
        return np.array(gscd.get_chelpg_charges(file))

    def symmetrize(mapping, charges):
        for val in np.unique(mapping):
            for idc in np.flatnonzero(mapping == val):
                yield val, np.mean(charges[idc])

    def neutralize(charges):
        return np.array(charges) - np.mean(charges)

    mol_mappings = enumerate(zip(mappings, entities), start=1)
    atomic_charges = {idx: chrg for mol_idx, (mapping, mol) in mol_mappings
                      for idx, chrg in symmetrize(mapping, read_charges(mol, mol_idx))}

    unitcell_charges = neutralize([
        atomic_charges[pos] for pos in equivalent_positions])

    np.savetxt("charges.dat", unitcell_charges)

    # Make molcas basis file for environment charges
    cells.write_molcas_basis(
        elements, dict(enumerate(unitcell_charges, start=1)), "ENV")

    print("\033[93m")
    print("Molcas basis set file for environment written to:")
    print("     ENV (basis set specification)\033[0m")


def cluster_func(args):

    structure = Structure.from_file(args.poscar)
    central_index = args.central_index

    if args.from_central:
        molecular_graph = cells.generate_molecular_graph(structure)
        connected_graphs = cells.split_molecular_graph(molecular_graph, filter_unique=False)
        central_graph = next(filter(lambda nodes: central_index - 1 in nodes, connected_graphs))
        _, qm_region = cells.extract_molecule(structure, central_graph, central_index=central_index - 1)
    elif args.qm_region is not None:
        _, qm_region = Molecule.from_file(args.qm_region)
    else:
        raise NotImplementedError()

    central_site = structure.sites[central_index - 1]
    shift = np.array([central_site.x, central_site.y, central_site.z])

    if args.distortion_expansion is not None:
        cluster = cells.DistortionSupercell.from_poscar(
            args.poscar, args.distortion_expansion, central_index=central_index)
    elif args.distortion_cutoff is not None:
        cluster = cells.DistortionCluster.from_poscar(
            args.poscar, args.distortion_cutoff, central_index=central_index)

    def format_label(args):
        idx, specie = args
        return f"{specie.symbol}.ENV.{idx}"

    labels = list(map(format_label, enumerate(structure.species, start=1))) * cluster.n_cell

    def generate_mapping(ref_coords):

        for idx, site in enumerate(qm_region.sites):

            coord = np.array([site.x, site.y, site.z]) - shift
            idc = np.flatnonzero(np.isclose(ref_coords, coord).all(axis=1))

            if len(idc) == 0:
                continue
            elif len(idc) == 1:
                yield idc[0]
            else:
                raise ValueError(f"Multiple instances of atom {coord}!")

    for idx in generate_mapping(cluster.cart_coords):
        labels[idx] = labels[idx].split('.')[0]

    xyzp.save_xyz("cluster.xyz", labels, cluster.cart_coords)


def derivatives_func(args, unknown_args=None, **func_kwargs):

    if args.method == "findiff":
        dx = Finite.from_file(args.distortion_info)

    # parse extra function kwargs
    if not func_kwargs and unknown_args:
        if args.function == 'CFPs':
            func_kwargs = vars(angmom_suite.read_args(['cfp'] + unknown_args))
        elif not unknown_args:
            func_kwargs = {}
        else:
            raise ValueError("Unused extra arguments not known to the parser.")

    func = make_func(args, max_numerical_order=dx.order,
                     max_analytic_order=args.max_analytic_order, **func_kwargs)

    # prints derivatives and function values on the input grid
    if args.format == 'points':
        dx.write_points(
            args.function + "_points.dat", func, max_order=args.order)
        return

    ho = Harmonic.from_file(args.vibration_info)

    # remove normalisation and convert to units of zero point displacement
    conv = np.sqrt(ho.red_masses) * (ho.zpd if args.zpd_unit else 1.0)
    dQ = dx.transform(ho.displacements, func, args.order) * conv[:, np.newaxis]

    if args.format == 'tau':  # print central derivatives in tau style input

        if not args.order == 1:
            raise ValueError("Only first order derivatives with tau format.")

        if not args.function == 'CFPs':
            raise ValueError("Only function=CFPs possible with tau format.")

        print_tau_style(dQ, ho.freqs, 'CFP_polynomials.dat')

    else:  # print central derivatives to file
        pass

    return


def generate_lvc_input_func(args):

    # todo: make prettier by parsing numbers directly to list of ints using
    # parse_range and a custom action which flattens, check also other uses of
    # parse_range if flattened

    # todo: replace dict by global definition
    roots_dict = {'dy': [18]}

    if args.num_roots:
        num_roots = map(int, args.num_roots)
    elif args.ion:
        num_roots = roots_dict[args.ion.lower()]
    else:
        sys.exit("Invalid specification of the number of roots.")

    for iph_idx, num_root in enumerate(num_roots, start=1):

        if num_root == 0:
            continue

        generate_lvc_input(
            args.old_path,
            args.old_project,
            num_root,
            iph_idx,
            mclr_extra=args.mclr_extra,
            alaska_extra=args.alaska_extra,
            two_step=args.two_step,
            dry=args.dry
        )

    return


def parse_range(string):
    """
    adapted from https://stackoverflow.com/questions/6512280/accept-a-range-of-
    numbers-in-the-form-of-0-5-using-pythons-argparse
    """

    # match numbers in hyphen separated range
    # capture first and second string of digits, don't capture hyphon
    splt = re.match(r'(\d+)(?:-(\d+))?$', string)

    if not splt:
        raise ArgumentTypeError(
            "'{}' is not a range of number. Expected forms like '0-5' or '2'.".format( # noqa
                string
            )
        )

    # extract start and end of range, if single number is given -> end = start
    start = splt.group(1)
    end = splt.group(2) or start

    return list(range(int(start), int(end)+1))


def strength_func(args):
    """
    Wrapper for spin-phonon coupling strength CLI call
    """

    # Load CFP polynomials from file
    dQ, freqs = read_tau_style(args.cfp_file)

    # TODO delete
    args.n = 13
    args.J = 3.5
    args.L = 3
    args.S = 0.5

    if not args.nooef:
        # Get OEF values
        OEFs = crystal.calc_oef(args.n, args.J, args.L, args.S)
        print(np.shape(OEFs))

        # Add back in OEFs to polynomial values
        dQ *= OEFs

    # Calculate strength values of each mode
    S = np.array([crystal.calc_total_strength(mode) for mode in dQ])

    # Save strength to file
    np.savetxt("strengths.dat", S)
    print("Strengths saved to strengths.dat")

    # Read in symmetry labels if given
    if args.irreps:
        irreps = list(np.loadtxt(args.irreps, dtype=str))
    else:
        irreps = ['A']*len(freqs)

    unique_irreps, irrep_ints = np.unique(irreps, return_inverse=True)    

    if args.plot:

        _, ax = plt.subplots(num='Spin-phonon coupling strength')

        for unique_int in np.unique(irrep_ints):
            ax.stem(
                freqs[np.nonzero(irrep_ints == unique_int)[0]],
                S[np.nonzero(irrep_ints == unique_int)[0]],
                basefmt=' ',
                linefmt='C{:d}-'.format(unique_int),
                markerfmt='C{:d}o'.format(unique_int)
            )
        if args.irreps:
            ax.legend(unique_irreps, frameon=False)

        ax.set_ylabel(r'$S$ (cm$^{-1}$)', fontname="Arial")
        ax.set_xlabel('Mode energy (cm$^{-1}$)', fontname="Arial")

        ax.set_ylim([0., np.max(S)*1.05])

        plt.savefig("strengths.svg")
        plt.savefig("strengths.png")
        print("Strength plots saved to strengths.svg and strengths.png")
        plt.show()

    return


cluster = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        add_help=False
)

morph = cluster.add_mutually_exclusive_group(required=True)

morph.add_argument(
    '--distortion_expansion',
    nargs=3,
    metavar=('N_x', 'N_y', 'N_z'),
    type=int,
    help='Supercell expansion.'
)

morph.add_argument(
    '--distortion_cutoff',
    type=float,
    help='Cut-off distance for unit cell cluster.'
)

cluster.add_argument(
    '--central_index',
    type=int,
    help='Index of the central spin center.'
)


def read_args(arg_list=None):
    description = '''
    A package for performing Spin-Phonon coupling calculations.
    '''

    parser = ArgumentParser(
            description=description,
            formatter_class=RawDescriptionHelpFormatter
            )

    subparsers = parser.add_subparsers(dest='prog')

    coords = subparsers.add_parser('coords')
    coords.set_defaults(func=coords_func)

    coords.add_argument(
        '-D',
        '--distortion_info',
        type=str,
        help='HDF5 database containing information about the distortion.'
    )

    coords.add_argument(
        '--num_atoms',
        type=int,
        default=1,
        help='Number of atoms.'
    )

    coords.add_argument(
        '--mode_wise',
        nargs='*',
        default=None,
        type=parse_range,
        help='Mode indices to be included in distortions calculations ' +
             '- activates mode-wise distortions.'
        )

    coords.add_argument(
        '--atomic',
        nargs='*',
        default=None,
        type=parse_range,
        help='Atomic indices to be included in distortion calculations ' +
             '- activates atomic distortions.'
        )

    coords.add_argument(
        '--num_steps',
        type=int,
        default=0,
        help='Number of distortion steps.'
    )

    coords.add_argument(
        '--order',
        type=int,
        default=1,
        help='Order of distortion.'
    )

    coords.add_argument(
        '--constant_step',
        type=float,
        help='Step size factor for constant displacement.'
    )

    evaluate = subparsers.add_parser('eval')
    evaluate.set_defaults(func=eval_func)

    evaluate.add_argument(
        '-H',
        '--Help',
        action=FunctionHelp,
        help='show help message for additional arguments and exit'
    )

    evaluate.add_argument(
        '-L',
        '--lvc_data',
        type=str,
        help='HDF5 database containing the LVC parameters.'
    )

    coords = evaluate.add_mutually_exclusive_group()

    coords.add_argument(
        '-V',
        '--vibration_info',
        type=str,
        help=('HDF5 database containing information about the vibrations.'
              'Derivatives are evaluated in the basis of dimension-less'
              'mass-frequency weighted normal mode coordinates.')
    )

    evaluate.add_argument(
        '--geom',
        nargs='+',
        help='*.xyz coordinates at which properties will be evaluated.'
    )

    evaluate.add_argument(
        '--order',
        type=int,
        default=0,
        help='Order of derivative.'
    )

    evaluate.add_argument(
        '--truncate',
        type=int,
        metavar="max_state",
        help='Truncate matrix elements at max_state'
    )

    evaluate.add_argument(
        '--align',
        action=BooleanOptionalAction,
        default=True,
        help='Align LVC model reference to input geometry by minimising RMSD.'
    )

    evaluate.add_argument(
        '--adiabatic_energies',
        nargs='+',
        action=OrderedSelectionAction,
        choices=['mch', 'soc'],
        help='Energies of the MCH or SOC eigenstates.'
    )

    evaluate.add_argument(
        '--diabatic_hamiltonian',
        nargs='+',
        action=OrderedSelectionAction,
        choices=['mch', 'soc'],
        help='Diabatic potential energy matrix between MCH or SOC states.'
    )

    evaluate.add_argument(
        '--proj',
        nargs='+',
        action=OrderedSelectionAction,
        help='Model Hamiltonian projection'
    )

    evaluate.add_argument(
        '--sus',
        nargs='+',
        action=OrderedSelectionAction,
        help='Magnetic susceptibility'
    )

    inp = subparsers.add_parser('generate_input')
    inp.set_defaults(func=generate_lvc_input_func)

    inp.add_argument(
        'old_project',
        type=str,
        help='Project name of preceding Molcas calculation.'
    )

    inp.add_argument(
        '--old_path',
        type=str,
        default='../',
        help='Path to WorkDir of preceding Molcas calculation.'
    )

    roots = inp.add_mutually_exclusive_group(required=True)

    roots.add_argument(
        '--num_roots',
        nargs='+',
        help='Number of states per JOBIPH.'
    )

    roots.add_argument(
        '--ion',
        type=str,
        help='Label of the metal center, e.g. Dy.'
    )

    inp.add_argument(
        '--jobiph',
        nargs='+',
        help='Indices of Molcas JOBIPH wavefunction files *_IPH.'
    )

    inp.add_argument(
        '--mclr_extra',
        nargs='+',
        default=((), {}),
        type=make_parse_dict(str, str, key_only=True),
        action=ParseExtra,
        help='Manually run mclr with custom options, e.g. thre=1e-8',
        metavar='name=value')

    inp.add_argument(
        '--alaska_extra',
        nargs='+',
        default=((), {}),
        type=make_parse_dict(str, str, key_only=True),
        action=ParseExtra,
        help='Run alaska with custom options, e.g. cuto=1e-8',
        metavar='name=value')

    inp.add_argument(
        '--two_step',
        action=BooleanOptionalAction,
        default=True,
        help='Utilize two-step procedure for MCLR runs'
    )

    inp.add_argument(
        '--dry',
        default=False,
        action='store_true',
        help='Dry-run which prints files to be created'
    )

    lvc = subparsers.add_parser('lvc')
    lvc.set_defaults(func=lvc_func)

    lvc.add_argument(
        '-L', '--lvc_data',
        type=str,
        help='HDF5 database output containing the LVC data.'
    )

    grad = lvc.add_mutually_exclusive_group()

    grad.add_argument(
        '--grad',
        type=str,
        nargs='+',
        help='Molcas output file(s) containing gradients and NACs.'
    )

    grad.add_argument(
        '--efld',
        type=str,
        nargs='+',
        help='Molcas output file(s) containing electric field values.'
    )

    grad.add_argument(
        '--add',
        type=str,
        nargs='+',
        help='LVC HDF5 data bases to be added.'
    )

    frozen = lvc.add_argument_group('Frozen density gradient arguments')

    frozen.add_argument(
        '--charge',
        type=str,
        nargs='+',
        help='Text file(s) containing point charge values.'
    )

    frozen.add_argument(
        '--xyz',
        type=str,
        nargs='+',
        help='xyz file(s) containing point charge coordinates.'
    )

    add = lvc.add_argument_group('LVC addition arguments')

    add.add_argument(
        '--reference_xyz',
        type=str,
        help='xyz file containing the reference coordinates.'
    )

    lvc.add_argument(
        '--rassi',
        type=str,
        help=('Molcas *.rassi.h5 output file containing AMFI integrals, '
              'SF_ANGMOM operators and the spin multiplicities.')
    )

    subset = lvc.add_mutually_exclusive_group()

    subset.add_argument(
        '--active_atoms',
        nargs='+',
        type=parse_range,
        help=(
            'Atomic indices active during spin-phonon coupling. Effectively '
            'effectively setting the coupling of all other atoms to zero. '
            'Useful to suppress coupling of specific atoms.'
        )
    )

    subset.add_argument(
        '--ref_coordinates',
        help='xyz coordinates containing the atomic positions of active atoms.'
    )

    lvc.add_argument(
        '--verbose',
        action='store_true',
        help='Plot gradient norm and invariance measures.'
    )

    build = subparsers.add_parser('cluster', parents=[cluster])
    build.set_defaults(func=cluster_func)

    build.add_argument(
        '--poscar',
        type=str,
        help='Unit cell POSCAR.'
    )

    qm = vib_calc_excl = build.add_mutually_exclusive_group(required=True)

    qm.add_argument(
        '--from_central',
        action='store_true',
        help=('QM region is built by completing the molecule around the '
              'central index.')
    )

    qm.add_argument(
        '--qm_region',
        type=str,
        help='Coordinates of the QM-region.'
    )

    vibrate = subparsers.add_parser('vib', parents=[cluster])
    vibrate.set_defaults(func=vib_func)

    vibrate.add_argument(
        '-V', '--vibration_info',
        type=str,
        help='HDF5 database containing information about the vibrations.'
    )

    vib_calc_excl = vibrate.add_mutually_exclusive_group(required=True)

    vib_calc_excl.add_argument(
        '--gaussian_log',
        type=str,
        help='Text output of a gaussian freq calculation.'
    )

    vib_calc_excl.add_argument(
        '--gaussian_fchk',
        type=str,
        help='Formatted checkpoint file of a gaussian freq calculation.'
    )

    vib_calc_excl.add_argument(
        '--force_sets',
        type=str,
        help='FORCE_SETS file from VASP-phonopy pre-process.'
    )

    vib_calc_excl.add_argument(
        '--phonopy_hdf5',
        type=str,
        help='Phonon database from phono3py calculation.'
    )

    vibrate.add_argument(
        '--poscar',
        type=str,
        help='Unit cell POSCAR from VASP-phonopy pre-process.'
    )

    vibrate.add_argument(
        '--force_expansion',
        nargs=3,
        metavar=('N_x', 'N_y', 'N_z'),
        default=(1, 1, 1),
        type=int,
        help='Supercell expansion used in phonon calculation'
    )

    qpoints = vibrate.add_mutually_exclusive_group()

    qpoints.add_argument(
        '--q_mesh',
        nargs=3,
        metavar=('N_x', 'N_y', 'N_z'),
        type=int,
        help='Mesh in q-space for phonon evaluation.'
    )

    vibrate.add_argument(
        '--gamma_center',
        action=BooleanOptionalAction,
        default=True,
        help='Center q-mesh around gamma point.'
    )

    vibrate.add_argument(
        '--mass',
        type=make_parse_dict(str_int, float),
        default={},
        nargs='+',
        action=ParseKwargs,
        help='Modify atomic masses for isotopic substitution.',
        metavar='atom_index=mass or element_symbol=mass'
    )

    vibrate.add_argument(
        '--trans',
        action=BooleanOptionalAction,
        default=True,
        help='Project out three rigid body translations.'
    )

    vibrate.add_argument(
        '--rot',
        action=BooleanOptionalAction,
        default=True,
        help='Project out three rigid body rotations.'
    )

    subset = vibrate.add_mutually_exclusive_group()

    subset.add_argument(
        '--active_atoms',
        nargs='+',
        type=parse_range,
        help=(
            'Atomic indices active during spin-phonon coupling. Effectively'
            ' subsets the displacement vectors. Useful if coupling is '
            'evaluated with a subset of the atoms present in the vibrational'
            ' calculation.'
        )
    )

    subset.add_argument(
        '--ref_coordinates',
        help='xyz coordinates containing the atomic positions of active atoms.'
    )

    vibrate.add_argument(
        '--save_pyvibms',
        type=str,
        help='optional pyvibms output for visualisation in PyMol.'
    )

    vibrate.add_argument(
        '--store_vecs',
        action=BooleanOptionalAction,
        default=True,
        help='Flag to disable expensive storage of normal mode displacements.'
    )

    prepare = subparsers.add_parser('prep')
    prepare.set_defaults(func=prepare_func)

    prepare.add_argument(
        'program',
        choices=['tau', 'tau_direct'],
        help='Program for which to prepare inputs.'
    )

    data = prepare.add_argument_group('database files')

    data.add_argument(
        '-L', '--lvc_data',
        help='HDF5 database containing the LVC parameters.')

    data.add_argument(
        '-V', '--vibration_info',
        help='HDF5 database containing information about the vibrations.')

    prepare.add_argument(
        '--ground_level',
        type=Level.parse,
        help='Symbol of the model space.'
    )

    prepare.add_argument(
        '--ion',
        type=Ion.parse,
        help='Central ion.'
    )

    prepare.add_argument(
        '--k_max',
        type=int,
        default=6,
        help='Maximum Stevens operator rank.'
    )

    prepare.add_argument(
        '--quax',
        action=angmom_suite.QuaxAction,
        help='Quantisation axes.'
    )

    prepare.add_argument(
        '--truncate',
        type=int,
        metavar="max_state",
        help='Truncate matrix elements at max_state'
    )

    prepare.add_argument(
        '--trace_less',
        action=BooleanOptionalAction,
        default=True,
        help='Substract out trace from bare coupling elements.'
    )

    charges = subparsers.add_parser(
        "charges",
        description="""
        Creates inputs for Gaussian CHELPG charge calculations of each
        entity in the unit cell of a VASP optimised structure and
        collects the resulting charges
        """)

    charges.set_defaults(func=charges_func)

    charges.add_argument(
        "poscar",
        type=str,
        help='Poscar containing optimised geometry'
    )

    mode = charges.add_mutually_exclusive_group(required=True)

    mode.add_argument(
        "--gen",
        action='store_true',
        help='Generate Gaussian ChelpG inputs.'
    )

    mode.add_argument(
        "--parse",
        action='store_true',
        help='Parse Gaussian ChelpG outputs.'
    )

    strength = subparsers.add_parser('strength')
    strength.set_defaults(func=strength_func)

    strength.add_argument(
        "cfp_file",
        type=str,
        help=(
            "File (hdf5 or CFP_polynomials) containing coupling crystal",
            "field parameters"
        )
    )

    strength.add_argument(
        "n",
        type=float,
        help=(
            "Number of unpaired electrons in 4f subshell"
        )
    )

    strength.add_argument(
        "J",
        type=float,
        help=(
            "Total angular momentum quantum number"
        )
    )

    strength.add_argument(
        "L",
        type=float,
        help=(
            "Orbital angular momentum quantum number"
        )
    )

    strength.add_argument(
        "S",
        type=float,
        help=(
            "Spin angular momentum quantum number"
        )
    )

    strength.add_argument(
        "--plot",
        action="store_true",
        help="Produce plot of strength as a function of mode energy"
    )

    strength.add_argument(
        '--irreps',
        type=str,
        metavar='<file_name>',
        help=(
            'Color code strength plot based on mode symmetries listed in file',
            'file must contain column of IRREPs, one per mode'
        )
    )

    strength.add_argument(
        "--nooef",
        action="store_true",
        help="Produce plot of strength as a function of mode energy"
    )

    # If arg_list==None, i.e. normal cli usage, parse_args() reads from
    # "sys.argv". The arg_list can be used to call the argparser from the
    # back end.

    # read sub-parser
    parser.set_defaults(func=lambda args: parser.print_help())
    _args, _ = parser.parse_known_args(arg_list)

    # select parsing option based on sub-parser
    if _args.prog in ['derivatives', 'eval']:
        args, hpc_args = parser.parse_known_args(arg_list)
        args.func(args, hpc_args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)


def main():
    read_args()



from operator import add, mul
from functools import reduce
from itertools import product
import numpy as np
import h5py
from ase.data import atomic_numbers
from pymatgen.core.structure import Structure
import phonopy
from phonopy.harmonic.dynmat_to_fc import get_commensurate_points

from gaussian_suite.extractor import make_extractor as make_gaussian_extractor

from .cells import build_cluster

C0 = 299.792458e6
BOHR2M = 5.29177210903e-11
AMU = 1.66053906660e-27
HARTREE2J = 4.3597447222071e-18
EV2J = 1.602176634e-19
H_PCK = 6.62607015E-34


class Harmonic:
    """Set of indenpendent quantum harmonic oscillators defined by their
    frequencies, displacements and reduced masses. The coordinate system
    follows the same conventions as the Gaussian software, i.e. the cartesian
    displacements are normalised and the normalisation constant is absorbed
    into the reduced mass.

    Parameters
    ----------
    freqs : np.array
        Array of harmonic frequencies in cm^-1.
    displacements : np.array
        K x N x 3 array containing the mass-weighted displacement vectors.

    Attributes
    ----------
    freqs : np.array
        Array of harmonic frequencies in units of cm^-1.
    displacements : np.array
        K x N x 3 array containing the mass-weighted displacement vectors.
    natoms : int
        Number of atoms.
    nmodes : int
        Number of modes.
    force_const : np.array
        Array of force constants in units of mdyne/ang or N/cm.
    zpd : np.array
        Array of zero point displacements in ang.
    """

    def __init__(self, freqs, displacements, coords=None,
                 weights=None, band_indices=None, q_points=None):

        self.nmodes = displacements.shape[0]
        self.natoms = displacements.shape[1]

        if freqs.shape[0] != self.nmodes:
            raise ValueError("Dimensions of HO parameters do not match.")

        self.freqs = freqs
        self.displacements = displacements

        self.coords = coords

        self.weights = weights
        self.band_indices = band_indices
        self.q_points = q_points

    @property
    def mass_freq_weighted_coordinates(self):
        """Mass-frequency weighted normal mode displacements. Equivalent to the
        conversion to zero-point displacement weighted coordinates divided by
        sqrt(2)."""
        conversion = 1.0e10 / np.sqrt(AMU) * np.sqrt(H_PCK / (4 * np.pi) / self.omega)
        return self.displacements * conversion[:, np.newaxis, np.newaxis]

    @property
    def omega(self):
        """angular frequency in s^-1"""
        return 2 * np.pi * (self.freqs * 100 * C0)

    def to_file(self, h_file, store_vecs=True):

        with h5py.File(h_file, 'w') as h:
            h['/'].attrs.create('num_modes', self.natoms)
            h['/'].attrs.create('num_atoms', self.natoms)
            h.create_dataset('frequencies', data=self.freqs)

            if store_vecs:
                h.create_dataset('displacements', data=self.displacements)

            if self.coords is not None:
                h.create_dataset('coords', data=self.coords)

            if self.weights is not None:
                h.create_dataset('weights', data=self.weights)

            if self.band_indices is not None:
                h.create_dataset('band_indices', data=self.band_indices)

            if self.q_points is not None:
                h.create_dataset('q_points', data=self.q_points)

    def to_pyvibms(self, file):
        with open(file, 'w') as f:
            f.write(f"{self.natoms} {self.nmodes}\n")
            for idx, (freq, displacement) in enumerate(zip(self.freqs, self.displacements), start=1):
                normalised = displacement.real / np.linalg.norm(displacement)
                f.write(f"\nN {freq:9.4f} A {idx}\n")
                f.write('\n'.join(f"{displ:9.4f}" for vec in normalised for displ in vec))
                f.write('\n')
            f.write("END\n")

    def subset_atoms(self, active_atom_idc=None, ref_coords=None):

        def generate_mapping():
            for idx, coord in enumerate(ref_coords):
                idc = np.flatnonzero(np.isclose(self.coords, coord).all(axis=1))
                if len(idc) == 0:
                    continue
                elif len(idc) == 1:
                    yield idc[0]
                else:
                    raise ValueError(f"Multiple instances of atom {coord}!")

        if active_atom_idc is not None and ref_coords is not None:
            raise ValueError("Supply one of active_atom_idc or ref_coords!")
        elif active_atom_idc is not None:
            idc = active_atom_idc
        elif ref_coords is not None:
            idc = list(generate_mapping())
        else:
            return self

        coords = None if self.coords is None else self.coords[idc]

        return self.__class__(self.freqs, self.displacements[:, idc], coords=coords)

    @classmethod
    def from_file(cls, h_file):

        with h5py.File(h_file, 'r') as h:
            displacements = h['displacements'][...]
            freqs = h['frequencies'][...]

            try:
                coords = h['coords'][...]
            except KeyError:
                coords = None

            try:
                weights = h['weights'][...]
            except KeyError:
                weights = None

            try:
                band_indices = h['band_indices'][...]
            except KeyError:
                band_indices = None

            try:
                q_points = h['q_points'][...]
            except KeyError:
                q_points = None

        return cls(freqs, displacements, coords=coords,
                   weights=weights, band_indices=band_indices,
                   q_points=q_points)

    @classmethod
    def from_gaussian_log(cls, f):
        freqs = make_gaussian_extractor(f, ("freq", "frequency"))[1]
        modes = make_gaussian_extractor(f, ("freq", "displacement"))[1]
        red_masses = make_gaussian_extractor(f, ("freq", "reduced_mass"))[1]
        return cls(freqs, modes / np.sqrt(red_masses)[:, np.newaxis, np.newaxis])

    @classmethod
    def from_gaussian_fchk(cls, f, ext_masses=None, **kwargs):
        hess = make_gaussian_extractor(f, ('fchk', 'hessian'))[()]
        masses = make_gaussian_extractor(f, ('fchk', 'atomic_mass'))[()]
        numbers = make_gaussian_extractor(f, ('fchk', 'atomic_number'))[()]
        coords = make_gaussian_extractor(f, ('fchk', 'coordinates'))[()]

        for key, mass in ext_masses.items():
            if isinstance(key, int):
                masses[key - 1] = mass
            elif isinstance(key, str):
                num = atomic_numbers[key]
                for idx in [i for i, n in enumerate(numbers) if n == num]:
                    masses[idx] = mass
            else:
                raise KeyError("Specify isotopic substitutions with either "
                               "element symbol (str) or atomic indices (int)!")

        return cls.analysis(hess, masses, coords=coords * BOHR2M * 1e10, **kwargs)

    @classmethod
    def generate_from_qpoints(cls, q_mesh, qpoint_list, freqs_list, vecs_list,
                              poscar, masses, trans=True, **cluster_kwargs):

        cluster = build_cluster(poscar, **cluster_kwargs)

        n_cell = cluster.n_cell

        cart = cluster.cart_coords
        frac = cluster.frac_coords

        # norm is sqrt(#qpoint)
        norm = np.sqrt(reduce(mul, q_mesh))

        for q_point, freqs, vecs in zip(qpoint_list, freqs_list, vecs_list):

            nmodes, natoms = vecs.shape[1], vecs.shape[0] // 3

            # reshape to K x N x 3 (K: #modes, N: #atoms)
            unitcell_displacements = vecs.T.reshape(nmodes, natoms, 3) / \
                np.sqrt(masses)[np.newaxis, :, np.newaxis] / norm

            if trans and not any(q_point):
                freqs = freqs[3:]
                unitcell_displacements = unitcell_displacements[3:]

            print(f"Evaluating q-point {q_point}")

            displacements = np.tile(unitcell_displacements, (1, n_cell, 1))

            if any(q_point):  # apply relative phases at off-gamma q-point
                phase = np.exp(2.j * np.pi * frac @ q_point)
                displacements *= phase[np.newaxis, :, np.newaxis]

            nmodes = freqs.size

            if is_boundary(q_point):

                abs_phase = np.array(list(map(phase_by_min_imag, displacements)))
                displacements *= abs_phase[:, np.newaxis, np.newaxis]

                imag_norm = np.linalg.norm(displacements.imag)
                real_norm = np.linalg.norm(displacements.real)
                print(f"{q_point}-point displacements have non-zero "
                      f"imaginary component: {imag_norm}, "
                      f"real component: {real_norm}!")

                yield cls(freqs, displacements.real, coords=cart,
                          band_indices=np.arange(1, nmodes + 1),
                          q_points=np.tile(q_point, (nmodes, 1)))
            else:

                yield cls(freqs, np.sqrt(2) * displacements.real, coords=cart,
                          band_indices=np.arange(1, nmodes + 1),
                          q_points=np.tile(q_point, (nmodes, 1)))

                yield cls(freqs, np.sqrt(2) * displacements.imag, coords=cart,
                          band_indices=np.arange(1, nmodes + 1),
                          q_points=np.tile(q_point, (nmodes, 1)))

    @classmethod
    def from_vasp_phonopy(cls, poscar, force_sets, ext_masses=None, trans=True,
                          force_expansion=(1, 1, 1), q_mesh=(1, 1, 1),
                          gamma_center=True, **cluster_kwargs):
        """Evaluate harmonic oscillators from VASP-phonopy calculation based
        on 
        """

        def eval_q(n):
            vals = np.linspace(0, 1, num=n, endpoint=False)
            center_val = vals[(n + 1) // 2 - 1]
            center_shift = np.mean(vals) if gamma_center else center_val
            return vals - center_shift

        q_points = list(filter(is_positive_q, product(*map(eval_q, q_mesh))))

        cell = phonopy.load(
            unitcell_filename=poscar,
            force_sets_filename=force_sets,
            supercell_matrix=force_expansion,
            primitive_matrix=np.identity(3)
        )

        masses = cell._dynamical_matrix._pcell.masses
        symbols = cell.unitcell.symbols

        for key, mass in ext_masses.items():
            if isinstance(key, int):
                masses[key - 1] = mass
            elif isinstance(key, str):
                for idx in [i for i, sym in enumerate(symbols) if sym == key]:
                    masses[idx] = mass
            else:
                raise KeyError("Specify isotopic substitutions with either "
                               "element symbol (str) or atomic indices (int)!")

        if ext_masses:
            cell._dynamical_matrix._pcell.masses = masses
            cell._set_dynamical_matrix()

        cell.run_qpoints(q_points, with_eigenvectors=True)
        qpoint_dict = cell.get_qpoints_dict()

        modes = Harmonic.generate_from_qpoints(
            q_mesh,
            q_points,
            qpoint_dict['frequencies'] * 1e12 / (C0 * 1e2),
            qpoint_dict['eigenvectors'],
            poscar,
            masses,
            trans=trans,
            **cluster_kwargs)

        return reduce(add, modes)

    @classmethod
    def from_phonopy_hdf5(cls, poscar, h_file, trans=True, **cluster_kwargs):

        with h5py.File(h_file, 'r') as h:

            q_mesh = h['mesh'][...]

            qpoint_list,  freqs_list, vecs_list = \
                zip(*filter(lambda args: is_positive_q(args[0]),
                            zip(h['grid_address'][...] / q_mesh,
                                h['frequency'][...] * 1e12 / (C0 * 1e2),
                                h['eigenvector'][...])))

        structure = Structure.from_file(poscar)
        masses = np.array([site.specie.atomic_mass for site in structure.sites])

        modes = Harmonic.generate_from_qpoints(
            q_mesh,
            qpoint_list,
            freqs_list,
            vecs_list,
            poscar,
            masses,
            trans=trans,
            **cluster_kwargs
        )

        return reduce(add, modes)

    def __add__(self, other):

        if not np.allclose(self.coords, other.coords):
            raise ValueError("Coordinates are incompatible!")

        order = np.argsort(np.concatenate((self.freqs, other.freqs)))

        def concat_order(attr):

            self_attr = getattr(self, attr)
            other_attr = getattr(other, attr)

            if self_attr is None or other_attr is None:
                return None
            else:
                return np.concatenate((self_attr, other_attr))[order]

        freqs = concat_order("freqs")
        displacements = concat_order("displacements")

        coords = self.coords

        weights = concat_order("weights")
        band_indices = concat_order("band_indices")
        q_points = concat_order("q_points")

        return self.__class__(freqs, displacements, coords=coords,
                              weights=weights, band_indices=band_indices,
                              q_points=q_points)

    @classmethod
    def analysis(cls, hessian, masses, trans=True, rot=True, coords=None):

        natom = len(masses)

        mwhess = hessian / np.sqrt(np.repeat(masses, 3)[np.newaxis, :] *
                                   np.repeat(masses, 3)[:, np.newaxis])

        eig, vec = np.linalg.eigh(mwhess)

        if trans:
            tra_frame = [np.array([d * np.sqrt(m) for m in masses for d in v])
                         for v in np.identity(3)]
            ntra = 3
        else:
            tra_frame = []
            ntra = 0

        if rot:
            # compute principle axis frame
            nrot, _, vec_inertia = principle_axis_inertia(masses, coords)

            # convert coordinates to principle axis frame
            # _coords = shift_to_com(masses, coords) @ vec_inertia.T
            _coords = shift_to_com(masses, coords)

            rot_frame = \
                [np.array([d * np.sqrt(m) for m, c in zip(masses, _coords)
                           for d in np.cross(c, v)])
                 for v in vec_inertia[:, (3 - nrot):].T]
        else:
            rot_frame = []
            nrot = 0

        nrotra = ntra + nrot
        nmodes = 3 * natom - nrotra

        rotra_frame = \
            np.array([d / np.linalg.norm(d) for d in tra_frame + rot_frame])

        if len(rotra_frame) != 0:
            # detect rigid body motions
            int_mask = np.logical_not(np.isclose(
                np.sum((rotra_frame @ vec)**2, axis=0), 1.0, rtol=1e-1))
            # Schmidt orthogonalisation
            new_frame, _ = np.linalg.qr(
                np.column_stack((rotra_frame.T, vec[:, int_mask])))
            # set coordinate frame to internal coordiantes
            frame = new_frame[:, nrotra:]

            # transfrom Hessian to internal coordinate frame and project out
            # rigid body motions
            eig, vec = np.linalg.eigh(frame.T @ mwhess @ frame)
            cart = frame @ vec / np.sqrt(np.repeat(masses, 3)[:, np.newaxis])

        else:
            cart = vec / np.sqrt(np.repeat(masses, 3)[:, np.newaxis])

        freqs = np.sqrt(eig * (HARTREE2J / BOHR2M**2 / AMU) /
                        (4*np.pi**2 * C0**2) + 0.j) / 100
        _freqs = np.vectorize(lambda x: -x.imag if x.imag else x.real)(freqs)

        displacements = np.reshape(cart.T, (nmodes, natom, 3))

        return cls(_freqs, displacements, coords=coords)


def is_positive_q(q):

    is_positive_or_gamma = (q[2] > 0 \
                            or (q[2] == 0 and q[1] > 0) \
                            or (q[2] == 0 and q[1] == 0 and q[0] > 0) \
                            or (q[2] == 0 and q[1] == 0 and q[0] == 0)) \
                           and not any(map(lambda x: x == -0.5, q))

    return is_positive_or_gamma


def is_boundary(q_point):  # Gamma point or boundary q-point predicate
    
    def on_boundary(q):
        return q == 0 or abs(q) == 0.5

    return all(map(on_boundary, q_point))


def phase_by_min_imag(x):

    ang = np.angle(x)
    mag = np.abs(x)**2

    phase = -np.arctan(np.sum(mag * np.sin(2 * ang)) / np.sum(mag * np.cos(2 * ang))) / 2

    if 2 * np.sum(mag * np.cos(2 * (ang + phase))) > 0:
        return np.exp(1.j * phase)
    else:
        return np.exp(1.j * (phase + np.pi / 2))


def shift_to_com(masses, coords):
    com = np.sum(masses[:, np.newaxis] * coords, axis=0) / np.sum(masses)
    return coords - com


def principle_axis_inertia(masses, coords):

    _coords = shift_to_com(masses, coords)

    inertia = np.zeros((3, 3))
    inertia[0, 0] = np.sum(masses * (_coords[:, 1]**2 + _coords[:, 2]**2))
    inertia[1, 1] = np.sum(masses * (_coords[:, 2]**2 + _coords[:, 0]**2))
    inertia[2, 2] = np.sum(masses * (_coords[:, 0]**2 + _coords[:, 1]**2))
    inertia[1, 0] = -np.sum(masses * (_coords[:, 0] * _coords[:, 1]))
    inertia[2, 0] = -np.sum(masses * (_coords[:, 0] * _coords[:, 2]))
    inertia[2, 1] = -np.sum(masses * (_coords[:, 1] * _coords[:, 2]))
    inertia[0, 1] = inertia[1, 0]
    inertia[0, 2] = inertia[2, 0]
    inertia[1, 2] = inertia[2, 1]

    eig_inertia, vec_inertia = np.linalg.eig(inertia)
    rank = np.linalg.matrix_rank(inertia)

    if rank > 1:
        return rank, eig_inertia, vec_inertia
    else:
        raise ValueError("Rank of moment of inertia tensor smaller than one.")




from fractions import Fraction
from functools import lru_cache, partial

import h5py

import numpy as np
from jax import jvp, jit
from jax.lax import stop_gradient
import jax.numpy as jnp
from jax.scipy.linalg import block_diag

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

from hpc_suite.store import Store
from molcas_suite.generate_input import MolcasComm, MolcasInput, Alaska, \
    Mclr, Emil
from molcas_suite.extractor import make_extractor as make_molcas_extractor
import angmom_suite
from angmom_suite.crystal import ProjectModelHamiltonian, \
    MagneticSusceptibility
from angmom_suite.basis import sf2ws, sf2ws_amfi, unitary_transform, sfy, \
    from_blocks, dissect_array, extract_blocks, make_angmom_ops_from_mult, \
    rotate_cart

ANG2BOHR = 1.88973


def extract_molcas_rassi(f_rassi):

    geom = make_molcas_extractor(f_rassi, ("rassi", "center_coordinates"))[()]

    spin_mult = make_molcas_extractor(f_rassi, ("rassi", "spin_mult"))[()]
    sf_mult = dict(zip(*np.unique(spin_mult, return_counts=True)))

    ener = make_molcas_extractor(f_rassi, ("rassi", "SFS_energies"))[()]
    sf_ener = list(dissect_array(ener, spin_mult))

    amfi = make_molcas_extractor(f_rassi, ("rassi", "SFS_AMFIint"))[()]
    sf_amfi = list(map(list, dissect_array(amfi, spin_mult, spin_mult)))

    angm = make_molcas_extractor(f_rassi, ("rassi", "SFS_angmom"))[()]
    sf_angm = list(extract_blocks(angm, spin_mult, spin_mult))

    return geom, sf_mult, sf_ener, sf_amfi, sf_angm


def extract_molcas_couplings(f_grad_list):

    for grad_file in f_grad_list:

        grad = make_molcas_extractor(grad_file, ("gradients", None))

        for (mult, root), val in iter(grad):
            yield mult, (root - 1, root - 1), val

        nacs = make_molcas_extractor(grad_file, ("nacs", "CI"))

        for (mult, root1, root2), val in iter(nacs):
            yield mult, (root1 - 1, root2 - 1), val


def extract_molcas_frozen_density_couplings(f_grad_list, charges):

    for grad_file, charge in zip(f_grad_list, charges):

        grad = make_molcas_extractor(grad_file, ("rasscf", "efld"))

        for (mult, root), val in iter(grad):
            yield mult, (root - 1, root - 1), -val * charge[:, np.newaxis]


class LVC:

    def __init__(self, sf_ener, sf_linear, sf_mult, geom, sf_amfi=None,
                 sf_angm=None):

        # LVC data
        self.sf_ener = sf_ener
        self.sf_linear = sf_linear

        # Spin multiplicities
        self.sf_mult = sf_mult

        self.sf_amfi = sf_amfi
        self.sf_angm = sf_angm

        self.geom = geom

    @property
    def natoms(self):
        return self.geom.shape[0]

    @property
    def ground_ener(self):
        return min(map(min, self.sf_ener))

    @property
    def sos_angm(self):
        return unitary_transform(sf2ws(self.sf_angm, self.sf_mult),
                                 self.diabatic_soc_trafo)

    @property
    def sos_spin(self):
        spin_mult = \
            [mult for mult, num in self.sf_mult.items() for _ in range(num)]
        spin = np.array(make_angmom_ops_from_mult(spin_mult)[0:3])
        return unitary_transform(spin, self.diabatic_soc_trafo)

    @property
    def diabatic_soc_trafo(self):
        mch = sf2ws(map(jnp.diag, self.sf_ener), self.sf_mult)
        soc = sf2ws_amfi(self.sf_amfi, self.sf_mult)
        _, vec = jnp.linalg.eigh(mch + soc)
        return vec

    def print_grad_norm(self):

        num = len(self.sf_mult)
        fig, axes = plt.subplots(nrows=num, ncols=1, squeeze=False)
        fig.set_size_inches(4, num * 4)

        for idx, (mult, lin) in enumerate(zip(self.sf_mult, self.sf_linear)):
            grad_norm = np.linalg.norm(lin, axis=(2, 3))
            axes[idx][0].title.set_text(f"Spin Multiplicity = {mult}")
            mat = axes[idx][0].matshow(grad_norm, norm=LogNorm())
            fig.colorbar(mat)

        plt.savefig("coupling_norm.pdf", dpi=600)

    def print_trans_rot_invariance(self):
        """Evaluates translational and rotational invariance along and around
        all three coordinate axes
        """

        num = len(self.sf_mult)
        fig, axes = plt.subplots(nrows=num, ncols=1, squeeze=False)
        nrm = np.sqrt(self.natoms)

        fig, axes = plt.subplots(nrows=num, ncols=3, squeeze=False)
        fig.set_size_inches(12, num * 4)

        for idx, (mult, lin) in enumerate(zip(self.sf_mult, self.sf_linear)):
            for vdx, (vec, lab) in enumerate(zip(np.identity(3), 'xyz')):
                inv = np.abs(np.einsum('ijkl,l->ij', lin, vec / nrm))
                inv[inv != 0] /= np.linalg.norm(lin, axis=(2, 3))[inv != 0]
                axes[idx][vdx].title.set_text(
                    f"axis = {lab}, Spin Multiplicity = {mult}")
                mat = axes[idx][vdx].matshow(inv, norm=LogNorm())
            fig.colorbar(mat, ax=axes[idx].tolist())

        plt.savefig("translational_invariance.pdf", dpi=600)

        fig, axes = plt.subplots(nrows=num, ncols=3, squeeze=False)
        fig.set_size_inches(12, num * 4)

        for idx, (mult, lin) in enumerate(zip(self.sf_mult, self.sf_linear)):
            for vdx, (ax, lab) in enumerate(zip(np.identity(3), 'xyz')):
                vec = np.array([np.cross(c, ax) for c in self.geom])
                inv = np.abs(np.einsum('ijkl,kl->ij', lin,
                                       vec / np.linalg.norm(vec)))
                inv[inv != 0] /= np.linalg.norm(lin, axis=(2, 3))[inv != 0]
                axes[idx][vdx].title.set_text(
                    f"axis = {lab}, Spin Multiplicity = {mult}")
                mat = axes[idx][vdx].matshow(inv, norm=LogNorm())
            fig.colorbar(mat, ax=axes[idx].tolist())

        plt.savefig("rotational_invariance.pdf", dpi=600)

    def to_file(self, f_lvc, verbose=False):

        if verbose:
            self.print_trans_rot_invariance()
            self.print_grad_norm()

        with h5py.File(f_lvc, 'w') as h:

            # LVC data
            for mult, ener in zip(self.sf_mult, self.sf_ener):
                h.create_dataset(f'{mult}/sf_ener', data=ener)

            for mult, lin in zip(self.sf_mult, self.sf_linear):
                h.create_dataset(f'{mult}/sf_linear', data=lin)

            # coordinates
            h.create_dataset('geom', data=self.geom)

            # spin multiplicities
            h.create_dataset('sf_mult', data=list(self.sf_mult.keys()))
            h.create_dataset('sf_nroots', data=list(self.sf_mult.values()))

            if self.sf_amfi is not None:
                for mult1, row in zip(self.sf_mult, self.sf_amfi):
                    for mult2, amfi in zip(self.sf_mult, row):
                        if abs(mult1 - mult2) <= 2:
                            h.create_dataset(f'{mult1}/{mult2}/sf_amfi',
                                             data=amfi)

            if self.sf_angm is not None:
                for mult, angm in zip(self.sf_mult, self.sf_angm):
                    h.create_dataset(f'{mult}/sf_angm', data=angm)

    @classmethod
    def from_file(cls, f_lvc):

        with h5py.File(f_lvc, 'r') as h:

            # coordinates
            geom = h['geom'][...]

            # Spin multiplicities
            sf_mult = dict(zip(h['sf_mult'][...], h['sf_nroots'][...]))

            # LVC data
            sf_ener = [h[f'{mult}/sf_ener'][...] for mult in sf_mult]
            sf_linear = [h[f'{mult}/sf_linear'][...] for mult in sf_mult]

            try:
                sf_amfi = [[h[f'{mult1}/{mult2}/sf_amfi'][...]
                            if abs(mult1 - mult2) <= 2 else
                            jnp.zeros((3, nroots1, nroots2))
                            for mult2, nroots2 in sf_mult.items()]
                           for mult1, nroots1 in sf_mult.items()]

            except KeyError:
                sf_amfi = None

            try:
                sf_angm = [h[f'{mult}/sf_angm'][...] for mult in sf_mult]
            except KeyError:
                sf_angm = None

        return cls(sf_ener, sf_linear, sf_mult, geom, sf_amfi=sf_amfi,
                   sf_angm=sf_angm)

    @classmethod
    def from_molcas_frozen_density(cls, f_rassi, f_grad_list, charge_list,
                                   coord_list):

        sf_linear_iter = \
            extract_molcas_frozen_density_couplings(f_grad_list, charge_list)
        _, sf_mult, sf_ener, sf_amfi, sf_angm = extract_molcas_rassi(f_rassi)
        geom = np.concatenate(coord_list)

        return cls.from_iter(sf_ener, sf_linear_iter, sf_mult, geom,
                             sf_amfi=sf_amfi, sf_angm=sf_angm)

    @classmethod
    def from_molcas(cls, f_rassi, f_grad_list, verbose=False):

        sf_linear_iter = extract_molcas_couplings(f_grad_list)
        geom, sf_mult, sf_ener, sf_amfi, sf_angm = extract_molcas_rassi(f_rassi)

        return cls.from_iter(sf_ener, sf_linear_iter, sf_mult, geom,
                             sf_amfi=sf_amfi, sf_angm=sf_angm)

    @classmethod
    def from_iter(cls, sf_ener, sf_linear_iter, sf_mult, geom,
                  sf_amfi=None, sf_angm=None):

        couplings = {mult: np.zeros((nroots, nroots) + geom.shape)
                     for mult, nroots in sf_mult.items()}

        for mult, (root1, root2), val in sf_linear_iter:
            couplings[mult][root1, root2, :] = val
            if root1 != root2:
                couplings[mult][root2, root1, :] = val

        return cls(sf_ener, list(couplings.values()), sf_mult, geom,
                   sf_amfi=sf_amfi, sf_angm=sf_angm)

    def join(self, other, ref_geom):

        if not (sfy(np.array_equal, sf=1)(self.sf_mult, other.sf_mult) and
                sfy(np.array_equal, sf=1)(self.sf_ener, other.sf_ener) and
                sfy(np.array_equal, sf=1)(self.sf_angm, other.sf_angm) and
                sfy(np.array_equal, sf=2)(self.sf_amfi, other.sf_amfi)):
            raise ValueError("LVC data not compatible!")

        sf_linear = map(lambda x, y: x + y, self.expand(ref_geom).sf_linear,
                        other.expand(ref_geom).sf_linear)

        return LVC(self.sf_ener, list(sf_linear), self.sf_mult, ref_geom,
                   sf_amfi=self.sf_amfi, sf_angm=self.sf_angm)

    def expand(self, ref_geom):

        def generate_mapping():
            for idx, coord in enumerate(self.geom):
                idc = np.flatnonzero(np.isclose(ref_geom, coord).all(axis=1))
                if len(idc) == 1:
                    yield (idx, idc[0])
                else:
                    raise ValueError(f"Multiple instances of atom {coord}!")

        atoms, mapping = zip(*generate_mapping())

        def expand(lin, num):
            zeros = jnp.zeros((num, num) + ref_geom.shape)
            return zeros.at[:, :, mapping].add(lin[:, :, atoms])

        sf_linear = map(expand, self.sf_linear, self.sf_mult.values())

        return LVC(self.sf_ener, list(sf_linear), self.sf_mult, ref_geom,
                   sf_amfi=self.sf_amfi, sf_angm=self.sf_angm)

    def subset_atoms(self, active_atom_idc=None, ref_coords=None):

        def generate_mapping():
            for idx, coord in enumerate(ref_coords):
                idc = np.flatnonzero(np.isclose(self.geom, coord).all(axis=1))
                if len(idc) == 0:
                    continue
                elif len(idc) == 1:
                    yield idc[0]
                else:
                    raise ValueError(f"Multiple instances of atom {coord}!")

        if active_atom_idc is not None and ref_coords is not None:
            raise ValueError("Supply one of active_atom_idc or ref_coords!")
        elif active_atom_idc is not None:
            idc = active_atom_idc
        elif ref_coords is not None:
            idc = list(generate_mapping())
        else:
            return self

        geom = self.geom[idc]
        sf_linear = [lin[:, :, idc, :] for lin in self.sf_linear]

        return self.__class__(self.sf_ener, sf_linear, self.sf_mult, geom,
                              sf_amfi=self.sf_amfi, sf_angm=self.sf_angm)

    def align(self, ref_geom):
        # translate centroid
        # find optimal rotation
        # according to http://nghiaho.com/?page_id=671

        def centroid(coords):
            return jnp.mean(coords, axis=1)

        self_centroid = centroid(self.geom)
        ref_centroid = centroid(ref_geom)

        uvec, _, vhvec = jnp.linalg.svd(
            (self.geom - self_centroid[:, jnp.newaxis]).T @
            (ref_geom - ref_centroid[:, jnp.newaxis]))

        rot = vhvec.T @ uvec.T

        if jnp.isclose(jnp.linalg.det(rot), -1.0):
            vhvec = vhvec.at[2].multiply(-1.0)

        rot = vhvec.T @ uvec.T

        geom = (self.geom - self_centroid[:, jnp.newaxis]) @ rot.T + \
            ref_centroid[:, jnp.newaxis]

        def rot_linear(y):
            return sfy(lambda x: jnp.einsum('ij,mnoj->mnoi', rot, x), sf=1)(y)

        sf_linear = rot_linear(self.sf_linear)
        sf_angm = sfy(lambda x: rotate_cart(x, rot), sf=1)(self.sf_angm)
        sf_amfi = sfy(lambda x: rotate_cart(x, rot), sf=2)(self.sf_amfi)

        return LVC(self.sf_ener, sf_linear, self.sf_mult, geom,
                   sf_amfi=sf_amfi, sf_angm=sf_angm)

    def diabatic_hamiltonian(self, geom, soc=False):

        distortion = geom - self.geom

        sf_mch = [jnp.diag(ener) + jnp.einsum('ijkl,kl->ij', lin, distortion)
                  for ener, lin in zip(self.sf_ener, self.sf_linear)]

        if soc:
            soc_potential = sf2ws_amfi(self.sf_amfi, self.sf_mult)
            soc_hamiltonian = sf2ws(sf_mch, self.sf_mult) + soc_potential
            return unitary_transform(soc_hamiltonian, self.diabatic_soc_trafo)
        else:
            return sf_mch

    def __call__(self, geom, soc=False):

        if soc:
            dia = self.diabatic_hamiltonian(geom, soc=True)
            soc_ener, soc_trafo = jnp.linalg.eigh(dia)

            def trafo(op):
                return unitary_transform(op, soc_trafo)

            return soc_ener, trafo

        else:
            dia = self.diabatic_hamiltonian(geom)
            sf_ener, sf_trafo = zip(*[jnp.linalg.eigh(v) for v in dia])

            def trafo(op, sf=0):
                Umat = sf_trafo if sf else from_blocks(*sf_trafo)
                return unitary_transform(op, Umat, sf=sf)

            return sf_ener, trafo


def make_lvc_evaluator(h_file, select, order=0, geom=None, coords=None,
                       truncate=None, align=True):

    lvc = LVC.from_file(h_file)

    if align and geom is not None:
        lvc = lvc.align(geom)

    item, option_str = select

    resolve_options = {
        "adiabatic_energies": lambda opt: {'soc': opt == "soc", 'truncate': truncate},
        "diabatic_hamiltonian": lambda opt: {'soc': opt == "soc", 'truncate': truncate},
        "proj": lambda opt: angmom_suite.read_args(['proj'] + opt.split()),
        "sus": lambda opt: angmom_suite.read_args(['sus'] + opt.split())
    }[item]

    options = resolve_options(option_str)

    evaluator = {
        "adiabatic_energies": LVCAdiabaticEnergies,
        "diabatic_hamiltonian": LVCDiabaticHamiltonian,
        "proj": LVCModelHamiltonian,
        "sus": LVCMagneticSusceptibility,
    }[item]

    return evaluator(lvc, order, geom=geom, coords=coords, **options)


class LVCData:

    def __init__(self, lvc, order, *args, geom=None, coords=None, **kwargs):
        self.lvc = lvc
        self.order = order
        self.geom = lvc.geom if geom is None else geom
        self.coords = np.identity(3 * lvc.natoms).reshape((-1, lvc.natoms, 3))\
            if coords is None else coords

        super().__init__(*args, **kwargs)

        self.meta['label'] = \
            tuple([f"coord_{i}" for i in range(order)]) + self.meta['label']

    def format_label(self, lab, axes):
        return axes + super().format_label(lab)

    def __iter__(self):

        def recursive_order(x, func, axes=(), order=0):

            if order == self.order:
                yield from map(lambda val, lab: (self.format_label(lab, axes), val), *func(x))

            elif order < self.order:
                first = axes[-1] if len(axes) else 0

                for axis, coord in enumerate(self.coords[first:], start=1):

                    def _func(x):
                        return jvp(func, (x,), (coord,), has_aux=True)[1:3]

                    yield from recursive_order(
                        x, _func, axes=axes + (axis,), order=order + 1)

        yield from recursive_order(self.geom, self.evaluate)


class LVCAdiabaticEnergies(LVCData, Store):
    def __init__(self, *args, soc=False, truncate=None, **kwargs):
        self.soc = soc
        self.truncate = truncate
        description = \
            ' '. join(["SOC" if self.soc else "MCH", "energies",
                       f"truncated to the lowest {self.truncate} states"
                       if self.truncate else ""])

        label = () if self.soc else ("multiplicity",)

        super().__init__(*args, "Energies", description, **kwargs, label=label,
                         units="cm^-1")

    def evaluate(self, x):

        if self.truncate is None or not self.soc:
            ener = self.lvc(x, soc=self.soc)[0]
        else:
            ener = self.lvc(x, soc=self.soc)[0][:self.truncate]

        if self.soc:
            return [ener], [()]
        else:
            return ener, map(lambda lab: (lab,), self.lvc.sf_mult)


class LVCDiabaticHamiltonian(LVCData, Store):
    def __init__(self, *args, soc=False, truncate=None, **kwargs):
        self.soc = soc
        self.truncate = truncate
        description = \
            ' '.join(["Diabatic Hamiltonian matrix between the",
                      f"lowest {self.truncate}" if self.truncate else "",
                      "SOC states" if soc else "MCH states"])

        super().__init__(*args, "Diabatic Hamiltonian", description, **kwargs,
                         label=(), units="cm^-1")

    def evaluate(self, x):

        if self.truncate is None or not self.soc:
            hamiltonian = self.lvc.diabatic_hamiltonian(x, soc=self.soc)
        else:
            idc = jnp.ix_(jnp.arange(self.truncate), jnp.arange(self.truncate))
            hamiltonian = self.lvc.diabatic_hamiltonian(x, soc=self.soc)[idc]

        if self.soc:
            return [hamiltonian], [()]
        else:
            return hamiltonian, map(lambda lab: (lab,), self.lvc.sf_mult)


class LVCModelHamiltonian(LVCData, ProjectModelHamiltonian):
    def __init__(self, lvc, *args, **kwargs):
        super().__init__(lvc, *args, None, lvc.sf_mult, **kwargs)

    def evaluate(self, x):

        # parse operators in diabatic basis
        ops = {'sf_angm': self.lvc.sf_angm,
               'sf_mch': self.lvc.diabatic_hamiltonian(x, soc=False),
               'sf_amfi': self.lvc.sf_amfi}

        return super().evaluate(**ops)


class LVCMagneticSusceptibility(LVCData, MagneticSusceptibility):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, None, **kwargs)

    def evaluate(self, x):

        # parse operators in diabatic basis

        return super().evaluate(
            hamiltonian=self.lvc.diabatic_hamiltonian(x, soc=True),
            spin=self.lvc.soc_spin,
            angm=self.lvc.soc_angm
        )


def reserve_space(string):
    return r"{:0" + str(len(str(string))) + r"}"


def generate_lvc_input(old_path, old_proj, num_root, jobiph_idx,
                       mclr_extra=((), {}), alaska_extra=((), {}),
                       two_step=True, dry=False):

    files = [
        Emil('copy', src, dest) for src, dest in [
            (f"{old_path}/{old_proj}.RunFile", "$Project.RunFile"),
            (f"{old_path}/{old_proj}.OneInt", "$Project.OneInt"),
            (f"{old_path}/{jobiph_idx}_IPH", "$Project.JobIph"),
            (f"{old_path}/{old_proj}.ChDiag", "$Project.ChDiag"),
            (f"{old_path}/{old_proj}.ChMap", "$Project.ChMap"),
            (f"{old_path}/{old_proj}.ChRed", "$Project.ChRed"),
            (f"{old_path}/{old_proj}.ChRst", "$Project.ChRst"),
            (f"{old_path}/{old_proj}.ChVec1", "$Project.ChVec1"),
            (f"{old_path}/{old_proj}.QVec00", "$Project.QVec00")
        ]
    ]

    root_pattern = reserve_space(num_root)
    input_name = f"lvc{jobiph_idx}_root{root_pattern}-{root_pattern}.input"

    if two_step:

        mclr_input = MolcasInput(
            *files, MolcasComm('MCLR first step'),
            Mclr(*mclr_extra[0], **mclr_extra[1], SALA=1, TWOStep="first"),
            title="LVC parametrisation generated by spin-phonon_suite"
        )

        if dry:
            print(f"lvc{jobiph_idx}_mclr.input")
        else:
            mclr_input.write(f"lvc{jobiph_idx}_mclr.input")

        mclr_files = [
            Emil('copy', src, dest) for src, dest in [
                (f"{old_path}/lvc{jobiph_idx}_mclr.tramo", "$Project.tramo"),
                (f"{old_path}/lvc{jobiph_idx}_mclr.Qdat", "$Project.Qdat")
            ]
        ]

        # set TWOStep flag for subsequent mclr runs and append extra file names
        mclr_extra[1]["TWOStep"] = "second"
        files.extend(mclr_files)

    for iroot in range(1, num_root + 1):

        grad_input = MolcasInput(
            *files, MolcasComm('Gradient'),
            Mclr(*mclr_extra[0], SALA=f"{iroot}", **mclr_extra[1]),
            Alaska(*alaska_extra[0], ROOT=f"{iroot}", **alaska_extra[1]),
            title="LVC parametrisation generated by spin-phonon_suite"
        )

        grad_file = input_name.format(iroot, iroot)

        if dry:
            print(grad_file)
        else:
            grad_input.write(grad_file)

        for jroot in range(iroot + 1, num_root + 1):

            nac_input = MolcasInput(
                *files, MolcasComm('Nonadiabatic coupling'),
                Mclr(*mclr_extra[0], NAC="{} {}".format(iroot, jroot),
                     **mclr_extra[1]),
                Alaska(*alaska_extra[0], NAC="{} {}".format(iroot, jroot),
                       **alaska_extra[1]),
                title="LVC parametrisation generated by spin-phonon_suite"
            )

            nac_file = input_name.format(iroot, jroot)

            if dry:
                print(nac_file)
            else:
                nac_input.write(nac_file)



import h5py
import numpy as np
from .vibrations import Harmonic

class CoordinateInfo:

    def __init__(self, method, natoms, ndisto, geom_list, distortion_idc,
                 distortion_mtx=None, order=1, nsteps=0, steps=None):

        self.method = method
        self.natoms = natoms
        self.ndisto = ndisto
        self.distortion_idc = distortion_idc
        self.order = order
        self.nsteps = nsteps
        self.steps = steps

        self._distortion_mtx = distortion_mtx

    @property
    def distortion_mtx(self):
        if self.method == 'atomic':
            return np.identity(3 * self.natoms).reshape((-1, self.natoms, 3))
        else:
            return self._distortion_mtx

    @classmethod
    def from_file(cls, h_file):

        with h5py.File(h_file, 'r') as h:
            method = h.attrs['method']
            natoms = h.attrs['num_atoms']
            ndisto = h.attrs['num_disto']
            distortion_idc = h['distortion_idc'][...]
            order = h.attrs['order']
            nsteps = h.attrs['num_steps']

            if nsteps > 0:
                steps = h['step_size'][...]
            else:
                steps = None

            try:
                distortion_mtx = h['distortion_mtx'][...]
            except KeyError:
                distortion_mtx = None

        return cls(method, natoms, ndisto, distortion_idc,
                   distortion_mtx=distortion_mtx, order=order,
                   nsteps=nsteps, steps=steps)

    @classmethod
    def from_args(cls, args):
        # parse distortion method
        method, ndisto, distortion_idc = resolve_distortion_method(args)

        if args.num_steps > 0:
            steps = resolve_step_sizes(args, args.num_atoms, distortion_idc)
        else:
            steps = None

        if method == 'mode_wise':
            distortion_mtx = \
                Harmonic.from_file(args.vibration_info).displacements
        else:
            distortion_mtx = None

        return cls(method, args.num_atoms, ndisto, distortion_idc,
                   distortion_mtx=distortion_mtx, order=args.order,
                   nsteps=args.num_steps, steps=steps)

    def to_file(self, h_file=None):

        with h5py.File(h_file, 'w') as h:
            h['/'].attrs.create('method', self.method)
            h['/'].attrs.create('num_atoms', self.natoms)
            h['/'].attrs.create('num_disto', self.ndisto)
            h.create_dataset('distortion_idc', data=self.distortion_idc)
            h['/'].attrs.create('order', self.order)
            h['/'].attrs.create('num_steps', self.nsteps)

            if self.steps is not None:
                h.create_dataset('step_size', data=self.steps)

            if self._distortion_mtx is not None:
                h.create_dataset('distortion_mtx', data=self._distortion_mtx)


def resolve_distortion_method(args):
    """Helper for the cli interface resolving the distortion method relevant
    arguments.
    """

    if args.mode_wise is not None:
        method = "mode_wise"

        distortion_idc = {int(m) - 1 for grp in args.mode_wise for m in grp}
        ndisto = len(distortion_idc)

    elif args.atomic is not None:
        method = "atomic"

        distortion_idc = {3 * (int(a) - 1) + c
                          for grp in args.atomic
                          for a in grp
                          for c in range(3)}
        ndisto = len(distortion_idc)

    else:
        raise ValueError("No distortion method specified.")

    return method, ndisto, list(distortion_idc)


def resolve_step_sizes(args, natoms, distortion_idc):

    step_list = [0.0 for _ in range(3 * natoms)]

    if args.constant_step:
        for idx in distortion_idc:
            step_list[idx] = args.constant_step

    else:
        print("Warning: Invalid step size specification. "
              "Only analytic derivatives.")

    return step_list



from itertools import product
from functools import reduce
from math import ceil

from pymatgen.io.vasp import Poscar
from pymatgen.core.structure import Molecule
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.local_env import JmolNN, CrystalNN, IsayevNN
import networkx as nx
import networkx.algorithms.isomorphism as iso

from phonopy.interface.vasp import read_vasp
import numpy as np
import numpy.linalg as la
import molvis.core as mvis
import matplotlib.cm as cm


def generate_molecular_graph(structure):
    strategy = IsayevNN(cutoff=3.0, allow_pathological=True)
    graph = StructureGraph.with_local_env_strategy(structure, strategy, weights=True)

    # build directed multi-graph
    twoway = nx.MultiDiGraph(graph.graph)
    # nodes are labelled by species
    nx.set_node_attributes(twoway, {node: {'specie': str(structure[node].specie)} for node in twoway})

    def flip_edge(u, v, data):
        return v, u, {key: tuple(-i for i in val) if key == 'to_jimage'
                      else val for key, val in data.items()}

    # edges are labelled by periodic image crossing
    twoway.add_edges_from([flip_edge(*edge) for edge in twoway.edges(data=True)])

    return twoway


def split_molecular_graph(graph, filter_unique=False):

    connected = nx.connected_components(nx.Graph(graph))

    subgraphs = map(graph.subgraph, connected)

    def filter_unique_graphs(graphs):

        edge_match = iso.numerical_edge_match("weight", 1.0)
        node_match = iso.categorical_node_match("specie", None)

        unique_graphs = set()

        for graph in graphs:

            def graph_match(g):
                return nx.is_isomorphic(
                    graph, g, node_match=node_match, edge_match=edge_match)

            already_present = map(graph_match, unique_graphs)

            if not any(already_present):
                unique_graphs.add(graph)
                yield graph

    if filter_unique:
        return list(filter_unique_graphs(subgraphs))
    else:
        return list(subgraphs)


def extract_molecule(structure, graph, central_index=None):

    # walk graph and consider to_jimage
    def generate_shifts():

        start = next(iter(graph.nodes())) if central_index is None else central_index

        def walk(shifts, edge):
            a, b = edge
            return shifts | {b: shifts[a] + graph[a][b][0]['to_jimage']}

        edges = nx.bfs_edges(graph, source=start)
        shifts = reduce(walk, edges, {start: np.zeros(3)})
        return dict(sorted(shifts.items()))

    shifts = generate_shifts()

    species = [structure.species[idx] for idx in shifts]
    coords = [structure.lattice.get_cartesian_coords(
        structure.frac_coords[idx] + shift) for idx, shift in shifts.items()]
    molecule = Molecule(species, coords)

    return list(shifts.keys()), molecule


def get_unique_entities(structure):

    molecular_graph = generate_molecular_graph(structure)
    connected_graphs = split_molecular_graph(molecular_graph, filter_unique=True)
    indices, molecules = zip(*map(lambda x: extract_molecule(structure, x), connected_graphs))

    sym_structure = SpacegroupAnalyzer(structure).get_symmetrized_structure()
    structure_dict = sym_structure.as_dict()
    equivalent_positions = structure_dict['equivalent_positions']

    elements = [site.specie.symbol for site in structure]

    mappings = [[equivalent_positions[idx] for idx in shift_dict]
                for shift_dict in indices]

    return equivalent_positions, elements, mappings, molecules


def build_cluster(poscar, central_index=None, distortion_expansion=None,
                  distortion_cutoff=None):

    if distortion_expansion is not None:
        cluster = DistortionSupercell.from_poscar(
            poscar, distortion_expansion, central_index=central_index)
    elif distortion_cutoff is not None:
        cluster = DistortionCluster.from_poscar(
            poscar, distortion_cutoff, central_index=central_index)

    return cluster


class DistortionExpansion:
    """
    Contains information on supercell created from a given cell with
    specified expansion

    Parameters
    ----------
    lat_vecs : list
        Lattice vectors of unit cell as rows
    frac_coords : list
        Coordinates of each atom in cell as fraction of cell parameters
        (spin centre at the origin)
    atom_numbers : list
        Atomic number of each atom in cell

    Attributes
    ----------
    expansion : list[int]
        Expansion used to generate supercell from cell e.g. [3, 3, 3]
    uc_sc_map : list
        Mapping of supercell atoms onto unit cell atoms
    """
    def __init__(self, lat_vecs, coords, atom_numbers, *args, **kwargs):

        self.lat_vecs = lat_vecs
        self.coords = self.recenter_unitcell(coords, *args)
        self.atom_numbers = atom_numbers

        self.cart_coords = self.generate_cluster(*args)
        self.frac_coords = self.cart_coords @ la.inv(self.lat_vecs)
        self.n_atoms = len(atom_numbers)
        self.n_cell = self.frac_coords.shape[0] // self.coords.shape[0]

    @classmethod
    def from_poscar(cls, poscar_name, *args, central_index=None):
        atoms = read_vasp(poscar_name)

        if central_index is None:
            coords = atoms.scaled_positions
        else:
            shift = atoms.scaled_positions[central_index - 1]
            coords = (atoms.scaled_positions - shift) % 1.

        return cls(atoms.cell, coords, atoms.numbers, *args)

    def generate_cluster(self, *args):

        def shift_coords(idc):
            return (self.coords + np.array(idc)) @ self.lat_vecs

        coords = np.array([coord for idc in self.generate_cell_idc(*args)
                           for coord in shift_coords(idc)])

        return self.recenter_cluster(coords, *args)


class DistortionSupercell(DistortionExpansion):

    def generate_cell_idc(self, expansion):

        def expansion_range(num):

            start = stop = num // 2

            if num % 2 == 0:  # even
                return range(-start, stop)
            elif num % 2 == 1:  # odd
                return range(-start, stop + 1)

        for nvec in product(*map(expansion_range, expansion)):
            yield nvec

    def recenter_cluster(self, cart, expansion):

        def shift(num, vec):
            if num % 2 == 0:  # even
                return 0.0
            elif num % 2 == 1:  # odd
                return vec / 2

        return cart - np.sum(list(map(shift, expansion, self.lat_vecs)), axis=0)

    def recenter_unitcell(self, frac, expansion):

        def shift(num):
            if num % 2 == 0:  # even
                return 0.0
            elif num % 2 == 1:  # odd
                return 1 / 2

        return (frac + np.array(list(map(shift, expansion)))) % 1.0


class DistortionCluster(DistortionExpansion):

    def generate_cell_idc(self, cutoff):

        def expansion_range(vec):
            num = ceil(cutoff / np.linalg.norm(vec))
            return range(-num, num + 1)

        for nvec in product(*map(expansion_range, self.lat_vecs)):

            r = np.sum([ni * ci for ni, ci in zip(nvec, self.lat_vecs)], axis=0)

            if np.linalg.norm(r) <= cutoff:
                yield nvec

    def recenter_cluster(self, cart, *args):
        return cart - np.sum(self.lat_vecs, axis=0) / 2

    def recenter_unitcell(self, frac, *args):
        return (frac + 1 / 2) % 1.0


def write_molcas_basis(labels, charge_dict, name):
    """
    Writes dummy molcas basis file for environment charges to a textfile named
    according to the basis name.

    Parameters
    ----------
    labels : list[str]
        Atomic labels of environment with no indexing
    charge_dict : dict
        CHELPG charge of each environment atom
    name : str
        Root name of basis
    """

    with open(name, 'w') as f:

        f.write("* This file was generated by spin_phonon_suite\n")
        for elem, (lab, chrg) in zip(labels, charge_dict.items()):
            f.write(f"/{elem}.{name}.{lab}.0s.0s.\n")
            f.write("Dummy basis set for atomic charges of environment\n")
            f.write("no ref\n")
            f.write(f"{chrg:.9f} 0\n")
            f.write("0 0\n")

    return


def vis_charges_viewer(coords, labels, norm_charges, extra_coords=[],
                       extra_labels=[], extra_color='', viewer_style_args={},
                       viewer_div_args={}, main_kwargs={}, extra_kwargs={}):

    ms1 = mvis.Model(labels, coords, **main_kwargs)

    colours = cm.get_cmap('coolwarm', 250)

    ms1.atom_colours = [
        '#{0:02x}{1:02x}{2:02x}'.format(*col[:-1])
        for col in colours(norm_charges, bytes=True)
    ]

    ms2 = mvis.Model(extra_labels, extra_coords, **extra_kwargs)

    if extra_color:
        ms2.atom_colours = extra_color

    viewer = mvis.Viewer(
        objects=[ms1, ms2],
        extra_div_args=viewer_div_args,
        extra_style_args=viewer_style_args
    )

    return viewer
